"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _path = _interopRequireDefault(require("path"));

var _worker_threads = require("worker_threads");

var _logger = _interopRequireDefault(require("@polkadot/util/logger"));

var _promisify = _interopRequireDefault(require("@polkadot/util/promisify"));

var _commands = _interopRequireDefault(require("./commands"));

// Copyright 2017-2018 @polkadot/client-db authors & contributors
// This software may be modified and distributed under the terms
// of the ISC license. See the LICENSE file for details.
// import u8aToHex from '@polkadot/util/u8a/toHex';
const l = (0, _logger.default)('db/main');
const returnable = ['get', 'root'];

class SyncDb {
  constructor() {
    (0, _defineProperty2.default)(this, "child", void 0);
    // NOTE With --experimental-modules (ESNext targets) this can become
    // this.child = new Worker(new URL('./worker.js', import.meta.url).pathname);
    this.child = new _worker_threads.Worker(_path.default.join(__dirname, './worker.js'));
  } // Sends a message to the worker, optionally (in the case of get/root) returning the
  // actual result.


  _sendMessage(type, key, value) {
    const state = new Int32Array(new SharedArrayBuffer(8));
    let view = null;
    let buffer = null;

    const start = () => {
      state[0] = _commands.default.START;
      this.child.postMessage({
        buffer,
        key,
        state,
        value,
        type
      });
      Atomics.wait(state, 0, _commands.default.START);
    }; // Ok, this is not something that returns a value, just send the message and
    // return when we the call has been done


    if (!returnable.includes(type)) {
      start();
      return null;
    } // The shared data buffer that will be used by the worker to send info back
    // (wrapped in Uint8Array for binary data or in DataView for Int32 data)


    const shared = new SharedArrayBuffer(4096);
    buffer = new Uint8Array(shared);
    view = new DataView(shared);
    let size = 0;
    let offset = 0;
    start(); // expect to read SIZE, END/ERROR here

    switch (state[0]) {
      case _commands.default.END:
      case _commands.default.ERROR:
        return null;
      // Ahah, we need to read the size (first result) to detemine how
      // big of a result buffer we need.

      case _commands.default.SIZE:
        size = view.getUint32(0);

        this._notifyFill(state);

        break;
      // This _should_ never happen... but...

      default:
        l.error('Unknown worker state', state[0]);
        return null;
    }

    const result = new Uint8Array(size); // Here we loop through the states and either read data to fill the buffer
    // or return when it is time to do so

    while (true) {
      switch (state[0]) {
        // we have reached the end, return what we have
        case _commands.default.END:
          return result;
        // Error, so just return a null for the caller to handle

        case _commands.default.ERROR:
          return null;
        // Get the available data from the buffer and write it into our result
        // array.

        case _commands.default.READ:
          const available = Math.min(buffer.length, size - offset);
          result.set(buffer.subarray(0, available), offset);
          offset += available;

          this._notifyFill(state);

          break;
        // This _should_ never happen... but...

        default:
          l.error('Unknown worker state', state[0]);
          return null;
      }
    }
  } // Notifies the worker that it should continue filling the result buffer


  _notifyFill(state) {
    state[0] = _commands.default.FILL; // FIXME This is going to be renamed '.notify', not in Node (yet)

    Atomics.wake(state, 0, 1);
    Atomics.wait(state, 0, _commands.default.FILL);
  }

  checkpoint() {
    this._sendMessage('checkpoint');
  }

  commit() {
    this._sendMessage('commit');
  }

  revert() {
    this._sendMessage('revert');
  }

  del(key) {
    this._sendMessage('del', key);
  }

  get(key) {
    return this._sendMessage('get', key);
  }

  put(key, value) {
    this._sendMessage('put', key, value);
  }

  trieRoot() {
    return this._sendMessage('root');
  }

  async terminate() {
    return (0, _promisify.default)(this.child, this.child.terminate);
  }

}

exports.default = SyncDb;